#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include <avr/pgmspace.h>
#include "usbdrv.h"

#define QSIZE 100
#define THRESHOLD 200
#define REPLACE_OLD_SAMPLE 0

typedef struct queue {
	int head;
	int tail;
	int16_t data[QSIZE];
} queue_t;

typedef struct report {
	uint8_t buttonMask;
	int8_t dx;
	int8_t dy;
	int8_t dWheel;
} report_t;

/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

PROGMEM const char usbHidReportDescriptor[52] = {
	/* USB report descriptor, size must match usbconfig.h */
	0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
	0x09, 0x02,                    // USAGE (Mouse)
	0xa1, 0x01,                    // COLLECTION (Application)
	0x09, 0x01,                    //   USAGE (Pointer)
	0xA1, 0x00,                    //   COLLECTION (Physical)
	0x05, 0x09,                    //     USAGE_PAGE (Button)
	0x19, 0x01,                    //     USAGE_MINIMUM
	0x29, 0x03,                    //     USAGE_MAXIMUM
	0x15, 0x00,                    //     LOGICAL_MINIMUM (0)
	0x25, 0x01,                    //     LOGICAL_MAXIMUM (1)
	0x95, 0x03,                    //     REPORT_COUNT (3)
	0x75, 0x01,                    //     REPORT_SIZE (1)
	0x81, 0x02,                    //     INPUT (Data,Var,Abs)
	0x95, 0x01,                    //     REPORT_COUNT (1)
	0x75, 0x05,                    //     REPORT_SIZE (5)
	0x81, 0x03,                    //     INPUT (Const,Var,Abs)
	0x05, 0x01,                    //     USAGE_PAGE (Generic Desktop)
	0x09, 0x30,                    //     USAGE (X)
	0x09, 0x31,                    //     USAGE (Y)
	0x09, 0x38,                    //     USAGE (Wheel)
	0x15, 0x81,                    //     LOGICAL_MINIMUM (-127)
	0x25, 0x7F,                    //     LOGICAL_MAXIMUM (127)
	0x75, 0x08,                    //     REPORT_SIZE (8)
	0x95, 0x03,                    //     REPORT_COUNT (3)
	0x81, 0x06,                    //     INPUT (Data,Var,Rel)
	0xC0,                          //   END_COLLECTION
	0xC0,                          // END COLLECTION
};

static volatile queue_t buffer;
static report_t report_buffer;
static const report_t zero = { .buttonMask = 0, .dx = 0, .dy = 0, .dWheel = 0 };
static uint8_t idleRate;

void queue_init(volatile queue_t *q)
{
	q->head = q->tail = 0;
}

int queue_add(volatile queue_t *q, int16_t value)
{
	int h = (q->head + 1) % QSIZE;
	if (h == q->tail) {
#if REPLACE_OLD_SAMPLE
		q->tail = (q->tail + 1) % QSIZE;
#else
		return -1;
#endif
	}
	q->data[q->head] = value;
	q->head = h;
	return 0;
}

int queue_get(volatile queue_t *q, int16_t *value)
{
	if (q->head == q->tail) {
		return -1;
	}
	*value = q->data[q->tail];
	q->tail = (q->tail + 1) % QSIZE;
	return 0;
}

ISR(ADC_vect)
{
	queue_add(&buffer, ADC);
}

void ADC_init(void)
{
	queue_init(&buffer);
	/** ADMUX settings
	 * MUX4..0 = 0..0       => ADC0 is selected
	 * ADLAR = 0            => bits are right adjusted
	 * REFS1 = 0, REFS0 = 0 => AVCC with external capacitor at AREF pin
	 */
	ADMUX = 0;
	ADMUX |= (1 << REFS0);
	/** ADCSRA settings
	 * Auto triggered
	 * Using interrupts
	 * ADPS2..0 = 1..1      => prescaler at 128
	 */
	ADCSRA = 0;
	ADCSRA |= (1 << ADEN);
	ADCSRA |= (1 << ADSC);
	ADCSRA |= (1 << ADATE);
	ADCSRA |= (1 << ADIE);
	ADCSRA |= (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);

	/* Activate interrupts */
	sei();
}

/* ------------------------------------------------------------------------- */

usbMsgLen_t usbFunctionSetup(uchar data[8])
{
	usbRequest_t	*rq = (void *)data;
	/* The following requests are never used. But since they are required by
	 * the specification, we implement them in this example.
	 */
	if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) {
		/* class request type */
		if (rq->bRequest == USBRQ_HID_GET_REPORT) {
			/* wValue: ReportType (highbyte), ReportID (lowbyte)
			 * we only have one report type, so don't look at
			 * wValue
			 */
			usbMsgPtr = (void *)&report_buffer;
			return sizeof(report_buffer);
		} else if (rq->bRequest == USBRQ_HID_GET_IDLE) {
			usbMsgPtr = &idleRate;
			return 1;
		} else if (rq->bRequest == USBRQ_HID_SET_IDLE) {
			idleRate = rq->wValue.bytes[1];
		}
	} else {
		/* no vendor specific requests implemented */
	}
	return 0; /* default for not implemented requests: return no data back
		   *to host
		   */
}

/* ------------------------------------------------------------------------- */

void mouse_init(void)
{
	int i = 0;
	wdt_enable(WDTO_1S);
	cli();
	usbInit();
	usbDeviceDisconnect();
	while (--i) {
		wdt_reset();
		_delay_ms(1);
	}
	usbDeviceConnect();
	sei();
}

void report_set_buttons(report_t *report, int16_t value)
{
	if (value > THRESHOLD) {
		report->buttonMask = 1;
	} else {
		report->buttonMask = 0;
	}
}

void main_loop(void)
{
	int16_t val;
	while (1) {
		report_buffer = zero;
		wdt_reset();
		usbPoll();
		if (usbInterruptIsReady()) {
			if (!queue_get(&buffer, &val)) {
				report_set_buttons(&report_buffer, val);
			}
			usbSetInterrupt((void *)&report_buffer,
					sizeof(report_t));
		}
	}
}

int main(void)
{
	mouse_init();
	ADC_init();
	main_loop();
	return 0;
}

